package org.kantega.jexmec.manager;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.*;

/**
 * Default implementation of PluginManager. Tracks the ClassLoaders registered by {@link org.kantega.jexmec.ClassLoaderProvider}s and uses the provided
 * {@link org.kantega.jexmec.PluginLoader} implementations with {@link org.kantega.jexmec.ServiceLocator}s to load plugins from the ClassLoaders.
 */
public final class DefaultPluginManager<P> implements PluginManager<P> {

    private final Class<P> pluginClass;

    private final List<PluginLoader<P>> pluginLoaders;

    private final List<ClassLoaderProvider> classLoaderProviders;

    private final ServiceLocator serviceLocator;

    private final EachListener eachListener;

    private final PluginRegistry registry;

    private volatile State state = State.UNSTARTED;

    /**
     * Creates a DefaultPluginManager managing the given type of plugins
     *
     * @param pluginClass                the class of the plugin to manage.
     * @param pluginLoaders              the collection of plugin loaders to use
     * @param classLoaderProviders the collection of plugin class loader providers to use
     * @param serviceLocators            the collection of service locators to use
     * @throws IllegalArgumentException if the plugin class is not an interface
     */
    private DefaultPluginManager(Class<P> pluginClass,
                                Collection<PluginLoader<P>> pluginLoaders,
                                Collection<ClassLoaderProvider> classLoaderProviders,
                                Collection<ServiceLocator> serviceLocators) {
        this.pluginClass = pluginClass;
        this.pluginLoaders = new ArrayList<PluginLoader<P>>(pluginLoaders);
        this.classLoaderProviders = new ArrayList<ClassLoaderProvider>(classLoaderProviders);
        serviceLocator = new CompoundServiceLocator(new ArrayList<ServiceLocator>(serviceLocators));
        eachListener = new EachListener();
        registry = new PluginRegistry();
    }

    /**
     * Returns a service locator capable of locating all services added to this plugin manager.
     *
     * @return the service locator
     */
    public ServiceLocator getServiceLocator() {
        return serviceLocator;
    }

    /**
     * Starts the plugin manager.
     *
     * @return the plugin manager
     */
    public synchronized DefaultPluginManager<P> start() {
        state.assertEquals(State.UNSTARTED, "starting plugin manager");

        eachListener.beforePluginManagerStarted(this);

        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.start();
        }

        ClassLoader parentClassLoader = new ResourceHidingClassLoader(getClass().getClassLoader(), pluginClass);

        for (ClassLoaderProvider provider : classLoaderProviders) {
            provider.start(new ProviderAwarePluginClassLoaderRegistry(provider, registry), parentClassLoader);
        }

        state = State.STARTED;

        eachListener.afterPluginManagerStarted(this);

        return this;
    }

    /**
     * Stops the plugin manager.
     *
     * @return the plugin manager.
     */
    public synchronized DefaultPluginManager<P> stop() {
        state.assertEquals(State.STARTED, "stopping plugin manager");

        eachListener.beforePluginManagerStopped(this);

        for (ClassLoaderProvider provider : classLoaderProviders) {
            provider.stop();
        }

        for (PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.stop();
        }

        state = State.STOPPED;

        eachListener.afterPluginManagerStopped(this);

        return this;
    }


    /**
     * Return a list of the plugins currently installed in this PluginManager
     *
     * @return the list of plugins
     */

    public List<P> getPlugins() {
        state.assertEquals(State.STARTED, "getting plugins. You need to call start() *before* calling getPlugins()");
        return Collections.unmodifiableList(registry.getPlugins());
    }

    @SuppressWarnings("unchecked")
    public <T extends P> List<T> getPlugins(Class<T> pluginClass) {
        if(pluginClass == this.pluginClass)
            return (List<T>) getPlugins();

        List<T> plugins = new ArrayList<T>();

        for(P plugin : getPlugins()) {
            if(pluginClass.isAssignableFrom(plugin.getClass())) {
                plugins.add((T) plugin);
            }
        }
        
        return plugins;
    }

    /**
     * Register a PluginManagerListener for event notification.
     *
     * @param listener the listener to add
     */
    public void addPluginManagerListener(PluginManagerListener<P> listener) {
        eachListener.addListener(listener);
    }

    /**
     * Unregister a PluginManagerListener for event notification.
     *
     * @param listener the listener to remove
     */
    public void removePluginManagerListener(PluginManagerListener<P> listener) {
        eachListener.removeListener(listener);
    }

    /**
     * Return the ClassLoader that initiated the loading of the given plugin. Note that this
     * need not be the same as <code>plugin.getClassLoader()</code>
     *
     * @param plugin the plugin for which to return the loading plugin
     * @return the initiating ClassLoader
     */
    public ClassLoader getClassLoader(P plugin) {
        return registry.getClassLoader(plugin);
    }

    /**
     * A registry keeping track of plugins.
     */
    private class PluginRegistry implements ClassLoaderProvider.Registry {

        private final Map<ClassLoader, Map<PluginLoader<P>, List<P>>> plugins;

        private final Map<P, ClassLoader> classLoaderMap;

        private PluginRegistry() {
            plugins = new LinkedHashMap<ClassLoader, Map<PluginLoader<P>, List<P>>>();
            classLoaderMap = new HashMap<P, ClassLoader>();
        }

        public <T extends ClassLoader> void add(Iterable<T> classLoaders) {
            replace(Collections.<T>emptySet(), classLoaders);
        }

        public <T extends ClassLoader> void remove(Iterable<T> loaders) {
            replace(loaders, Collections.<T>emptySet());
        }

        public <T extends ClassLoader> void replace(final Iterable<T> removeClassLoaders, Iterable<T> addClassLoaders) {

            removeClassLoaders(removeClassLoaders);

            addClassLoaders(addClassLoaders);

            eachListener.pluginsUpdated(getPlugins());
        }

        private <T extends ClassLoader> void removeClassLoaders(Iterable<T> removeClassLoaders) {

            ClassLoaderProvider provider = ProviderAwarePluginClassLoaderRegistry.getCurrentProvider();

            // Validate that all class loaders to remove are actually registered
            synchronized (this) {
                for (ClassLoader loader : removeClassLoaders) {
                    if (!plugins.containsKey(loader)) {
                        throw new IllegalArgumentException("Illegal request to remove a class loader that isn't currently registered: " + loader);
                    }
                }
            }

            for (ClassLoader loader : removeClassLoaders) {

                eachListener.beforeClassLoaderRemoved(DefaultPluginManager.this, provider, loader);

                Map<PluginLoader<P>, List<P>> pluginsRemoved;
                synchronized (this) {
                    pluginsRemoved = plugins.get(loader);
                }

                for (PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                    eachListener.beforePassivation(DefaultPluginManager.this, provider, loader, pluginLoader, pluginsRemoved.get(pluginLoader));
                }

                synchronized (this) {
                    for (PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                        for (P plugin : pluginsRemoved.get(pluginLoader)) {
                            classLoaderMap.remove(plugin);
                        }
                    }
                    plugins.remove(loader);

                }

                for (PluginLoader<P> pluginLoader : pluginsRemoved.keySet()) {
                    eachListener.afterPassivation(DefaultPluginManager.this, provider, loader, pluginLoader, pluginsRemoved.get(pluginLoader));
                }


                for (PluginLoader<P> pluginLoader : pluginLoaders) {
                    pluginLoader.unloadPlugins(loader);
                }

                eachListener.afterClassLoaderRemoved(DefaultPluginManager.this, provider, loader);

            }
        }

        private <T extends ClassLoader> void addClassLoaders(Iterable<T> addClassLoaders) {
            ClassLoaderProvider provider = ProviderAwarePluginClassLoaderRegistry.getCurrentProvider();


            for (ClassLoader classLoader : addClassLoaders) {

                eachListener.beforeClassLoaderAdded(DefaultPluginManager.this, provider, classLoader);

                Map<PluginLoader<P>, List<P>> loadedPlugins = new LinkedHashMap<PluginLoader<P>, List<P>>();

                for (PluginLoader<P> pluginLoader : pluginLoaders) {
                    try {
                        loadedPlugins.put(pluginLoader, pluginLoader.loadPlugins(pluginClass, classLoader, serviceLocator));
                    } catch (Throwable e) {
                        eachListener.pluginLoadingFailedWithException(DefaultPluginManager.this, provider, classLoader, pluginLoader, e);
                    }
                }

                for (PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                    if (!loadedPlugins.get(pluginLoader).isEmpty()) {
                        eachListener.beforeActivation(DefaultPluginManager.this, provider, classLoader, pluginLoader, loadedPlugins.get(pluginLoader));
                    }
                }

                List<P> plugins = new ArrayList<P>();

                for (PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                    plugins.addAll(loadedPlugins.get(pluginLoader));
                }

                synchronized (this) {

                    for (P plugin : plugins) {
                        classLoaderMap.put(plugin, classLoader);
                    }

                    this.plugins.put(classLoader, loadedPlugins);
                }

                for (PluginLoader<P> pluginLoader : loadedPlugins.keySet()) {
                    if (!loadedPlugins.get(pluginLoader).isEmpty()) {
                        eachListener.afterActivation(DefaultPluginManager.this, provider, classLoader, pluginLoader, loadedPlugins.get(pluginLoader));
                    }
                }

                eachListener.afterClassLoaderAdded(DefaultPluginManager.this, provider, classLoader);

            }
        }

        /**
         * Return currently installed plugins
         *
         * @return a list of installed plugins
         */
        public synchronized List<P> getPlugins() {
            List<P> plugins = new ArrayList<P>();

            for (Map<PluginLoader<P>, List<P>> map : this.plugins.values()) {
                for (PluginLoader<P> pluginLoader : map.keySet()) {
                    plugins.addAll(map.get(pluginLoader));
                }
            }
            return plugins;
        }

        /**
         * Returns the classloader that caused the loading of the given plugin.
         *
         * @param plugin the plugin
         * @return the classloader which the plugin was loaded from
         */
        public synchronized ClassLoader getClassLoader(P plugin) {
            return classLoaderMap.get(plugin);
        }

    }

    /**
     * A service locator that wraps a collecton of service locators. Services are looked up in
     * the collection order. The first non null service is returned. If no services are found, service
     * lookup returns null.
     */
    private class CompoundServiceLocator implements ServiceLocator {
        private final Collection<ServiceLocator> serviceLocators;

        CompoundServiceLocator(Collection<ServiceLocator> serviceLocators) {
            this.serviceLocators = serviceLocators;
        }

        /**
         * Gets the the set of service keys supported by the collection
         * of service locators.
         *
         * @return the combined set of service keys
         */
        public Set<ServiceKey> keySet() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            for (ServiceLocator serviceLocator : serviceLocators) {
                serviceKeys.addAll(serviceLocator.keySet());
            }
            return serviceKeys;

        }

        /**
         * Returns the first non-null service implementation matching the given key. If no service is
         * found, null is returned.
         *
         * @param serviceKey the service key class to find an implementation for
         * @param <T>        the type of the service
         * @return the service instance or null if the service was not found.
         */
        public <T> T get(ServiceKey<T> serviceKey) {
            for (ServiceLocator serviceLocator : serviceLocators) {
                if(serviceLocator.keySet().contains(serviceKey)) {
                    return serviceLocator.get(serviceKey);
                }
            }
            return null;
        }
    }

    static class ResourceHidingClassLoader extends ClassLoader {

        private final String[] localResourcePrefixes;

        /**
         * Creates a ResourceHidingClassLoader hiding resources in <code>META-INF/services/PluginName/</code> and
         * <code>META-INF/services/com.example.PluginName/</code>.
         *
         * @param parent      the parent class loader
         * @param pluginClass the plugin class to hide resources for.
         */
        public ResourceHidingClassLoader(ClassLoader parent, Class pluginClass) {
            super(parent);
            localResourcePrefixes = new String[]{"META-INF/services/" + pluginClass.getSimpleName() + "/",
                    "META-INF/services/" + pluginClass.getName() + "/"};
        }

        @Override
        public InputStream getResourceAsStream(String name) {
            final URL resource = getResource(name);
            try {
                return resource == null ? null : resource.openStream();
            } catch (IOException e) {
                return null;
            }
        }

        @Override
        public URL getResource(String name) {
            if (isLocalResource(name)) {
                return super.findResource(name);
            } else {
                return super.getResource(name);
            }
        }

        @Override
        public Enumeration<URL> getResources(String name) throws IOException {
            if (isLocalResource(name)) {
                return super.findResources(name);
            } else {
                return super.getResources(name);
            }
        }

        protected boolean isLocalResource(String name) {
            for (String localResourcePrefix : localResourcePrefixes) {
                if (name.startsWith(localResourcePrefix)) {
                    return true;
                }
            }
            return false;
        }
    }

    private static class ProviderAwarePluginClassLoaderRegistry implements ClassLoaderProvider.Registry {
        private static final ThreadLocal<ClassLoaderProvider> providerThreadLocal = new ThreadLocal<ClassLoaderProvider>();
        private final ClassLoaderProvider wrappedProvider;
        private final ClassLoaderProvider.Registry wrappedRegistry;

        public ProviderAwarePluginClassLoaderRegistry(ClassLoaderProvider wrappedProvider, ClassLoaderProvider.Registry wrappedRegistry) {
            this.wrappedProvider = wrappedProvider;
            this.wrappedRegistry = wrappedRegistry;
        }

        public <T extends ClassLoader> void add(Iterable<T> classLoaders) {
            providerThreadLocal.set(wrappedProvider);
            try {
                wrappedRegistry.add(classLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public <T extends ClassLoader> void remove(Iterable<T> classLoaders) {
            providerThreadLocal.set(wrappedProvider);
            try {
                wrappedRegistry.remove(classLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public <T extends ClassLoader> void replace(Iterable<T> removeClassLoaders, Iterable<T> addClassLoaders) {
            try {
                providerThreadLocal.set(wrappedProvider);
                wrappedRegistry.replace(removeClassLoaders, addClassLoaders);
            } finally {
                providerThreadLocal.remove();
            }
        }

        public static ClassLoaderProvider getCurrentProvider() {
            return providerThreadLocal.get();
        }
    }

    private enum State {
        UNSTARTED, STARTED, STOPPED;

        /**
         * Throws an IllegalStateException if the <code>actual<code> state is different from the <code>expected</code> state.
         *
         * @param expected the expected state
         * @param when     description of when this state assertion failed
         */
        public void assertEquals(State expected, String when) {
            if (this != expected) {
                throw new IllegalStateException("Illegal state " + this + ". Expected " + expected + " when " + when);
            }
        }
    }

    /**
     * Plugin listener which makes a synchronized dispatch of each of the event methods to a collection
     * of plugin listeners.
     */
    private class EachListener extends PluginManagerListener<P> {
        private final List<PluginManagerListener<P>> listeners;

        EachListener() {
            this.listeners = new ArrayList<PluginManagerListener<P>>();
        }

        @Override
        public synchronized void afterActivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Collection<P> plugins) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterActivation(pluginManager, classLoaderProvider, classLoader, pluginLoader, plugins);
        }

        @Override
        public synchronized void afterClassLoaderAdded(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterClassLoaderAdded(pluginManager, classLoaderProvider, classLoader);
        }

        @Override
        public synchronized void afterClassLoaderRemoved(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterClassLoaderRemoved(pluginManager, classLoaderProvider, classLoader);
        }

        @Override
        public synchronized void afterPassivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Collection<P> plugins) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterPassivation(pluginManager, classLoaderProvider, classLoader, pluginLoader, plugins);
        }

        @Override
        public synchronized void afterPluginManagerStarted(PluginManager pluginManager) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterPluginManagerStarted(pluginManager);
        }

        @Override
        public synchronized void afterPluginManagerStopped(PluginManager pluginManager) {
            for(PluginManagerListener<P> listener : listeners)
                listener.afterPluginManagerStopped(pluginManager);
        }

        @Override
        public synchronized void beforeActivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Collection<P> plugins) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforeActivation(pluginManager, classLoaderProvider, classLoader, pluginLoader, plugins);
        }

        @Override
        public synchronized void beforeClassLoaderAdded(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforeClassLoaderAdded(pluginManager, classLoaderProvider, classLoader);
        }

        @Override
        public synchronized void beforeClassLoaderRemoved(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforeClassLoaderRemoved(pluginManager, classLoaderProvider, classLoader);
        }

        @Override
        public synchronized void beforePassivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Collection<P> plugins) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforePassivation(pluginManager, classLoaderProvider, classLoader, pluginLoader, plugins);
        }

        @Override
        public synchronized void beforePluginManagerStarted(PluginManager<P> pluginManager) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforePluginManagerStarted(pluginManager);
        }

        @Override
        public synchronized void beforePluginManagerStopped(PluginManager<P> pluginManager) {
            for(PluginManagerListener<P> listener : listeners)
                listener.beforePluginManagerStopped(pluginManager);
        }

        @Override
        public synchronized void pluginLoadingFailedWithException(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Throwable exception) {
            for(PluginManagerListener<P> listener : listeners)
                listener.pluginLoadingFailedWithException(pluginManager, classLoaderProvider, classLoader, pluginLoader, exception);
        }

        @Override
        public synchronized void pluginsUpdated(Collection<P> plugins) {
            for (PluginManagerListener<P> listener : listeners)
                listener.pluginsUpdated(plugins);
        }

        public synchronized void addListener(PluginManagerListener<P> listener) {
            listeners.add(0, listener);
        }

        public synchronized void removeListener(PluginManagerListener<P> listener) {
            listeners.remove(listener);
        }
    }

    public static <P> Builder<P> buildFor(Class<P> pluginClass) {
        if(pluginClass == null) {
            throw new NullPointerException("Null passes as argument to buildFor");
        }
        return new Builder<P>(pluginClass);
    }

    /**
     * Provides a builder for DefaultPluginManagers
     * @param <P>
     */
    public static class Builder<P> {

        private final Class<P> pluginClass;

        private final List<PluginLoader<P>> pluginLoaders = new ArrayList<PluginLoader<P>>();

        private final List<ClassLoaderProvider> classLoaderProviders = new ArrayList<ClassLoaderProvider>();

        private final List<ServiceLocator> serviceLocators = new ArrayList<ServiceLocator>();

        private final List<PluginManagerListener<P>> pluginManagerListeners = new ArrayList<PluginManagerListener<P>>();

        private StaticServiceLocator pluginManagerService;

        private boolean concretePluginClassAllowed = false;
        private boolean pluginLoadingExceptionsSwallowed = false;

        /**
         * Constructor is private, use {@link DefaultPluginManager#buildFor(Class)}}
         * @param pluginClass the class to build a plugin manager for
         */
        private Builder(Class<P> pluginClass) {
            this.pluginClass = pluginClass;
        }

        public Builder<P> withPluginLoader(PluginLoader pluginLoader) {
            return withPluginLoaders(pluginLoader);
        }
        public Builder<P> withPluginLoaders(PluginLoader<P>... pluginLoaders) {
            this.pluginLoaders.addAll(Arrays.asList(pluginLoaders));
            return this;
        }

        public Builder<P> withClassLoader(ClassLoader classLoader) {
            withClassLoaders(classLoader);
            return this;
        }
        public Builder<P> withClassLoaders(ClassLoader... classLoaders) {
            this.classLoaderProviders.add(new StaticClassLoaderProvider(classLoaders));
            return this;
        }

        public Builder<P> withClassLoaderProvider(ClassLoaderProvider classLoaderProvider) {
            return withClassLoaderProviders(classLoaderProvider);
        }

        public Builder<P> withClassLoaderProviders(ClassLoaderProvider... classLoaderProviders) {
            this.classLoaderProviders.addAll(Arrays.asList(classLoaderProviders));
            return this;
        }

        /**
         * Adds a single service mapping. Shortcut for calling addServiceLocator(new StaticServiceLocator(ServiceKey, Object))
         *
         * @param serviceKey the key to add a service for
         * @param service    the service implementation
         * @return the plugin manager.
         */
        public <T> Builder<P> withService(ServiceKey<T> serviceKey, T service) {
            withServiceLocators(new StaticServiceLocator(serviceKey, service));
            return this;
        }

        public Builder<P> withServiceLocators(ServiceLocator... serviceLocators) {
            this.serviceLocators.addAll(Arrays.asList(serviceLocators));
            return this;
        }

        public Builder<P> withServiceLocator(ServiceLocator serviceLocators) {
            this.serviceLocators.addAll(Arrays.asList(serviceLocators));
            return this;
        }

        public Builder<P> withPluginLoadingExceptionsSwallowed() {
            this.pluginLoadingExceptionsSwallowed = true;
            return this;
        }

        public Builder<P> withConcretePluginClassAllowed() {
            this.concretePluginClassAllowed = true;
            return this;
        }
        public <T> Builder<P> withPluginManagerAsService(ServiceKey<T> pluginManagerServiceKey) {
            withServiceLocators(pluginManagerService = new StaticServiceLocator(pluginManagerServiceKey, null));
            return this;
        }

        public DefaultPluginManager<P> build() {
            if (!concretePluginClassAllowed && !pluginClass.isInterface()) {
                throw new IllegalArgumentException("Plugin class needs to be an interface: " + pluginClass);
            }
            DefaultPluginManager<P> manager = new DefaultPluginManager<P>(pluginClass, pluginLoaders, classLoaderProviders, serviceLocators);
            if(pluginManagerListeners != null) {
                if(!pluginLoadingExceptionsSwallowed) {
                    pluginManagerListeners.add(new RethrowPluginLoadingExceptionsListener());
                }
                for(PluginManagerListener<P> listener : pluginManagerListeners) {
                    manager.addPluginManagerListener(listener);
                }
            }
            if(pluginManagerService != null) {
                pluginManagerService.setService(manager);
            }
            return manager;
        }

        /**
         * Class loader provider backed by a static list of class loaders.
         */
        private static class StaticClassLoaderProvider implements ClassLoaderProvider {
            private final List<ClassLoader> classLoaders = new ArrayList<ClassLoader>();
            private Registry registry;

            public StaticClassLoaderProvider(ClassLoader... classLoaders) {
                this.classLoaders.addAll(Arrays.asList(classLoaders));
            }

            public void start(Registry registry, ClassLoader parentClassLoader) {
                this.registry = registry;
                registry.add(classLoaders);
            }

            public void stop() {
                registry.remove(classLoaders);
            }
        }

        /**
         * Service locator backed by a static service key and implementation.
         */
        private static class StaticServiceLocator implements ServiceLocator {

            private final Set<ServiceKey> serviceKeys;
            private volatile Object service;

            public StaticServiceLocator(ServiceKey serviceKey, Object service) {
                this.serviceKeys = Collections.singleton(serviceKey);
                this.service = service;
            }

            public Set<ServiceKey> keySet() {
                return serviceKeys;
            }

            public <T> T get(ServiceKey<T> serviceKey) {
                return serviceKey.getType().cast(service);
            }

            public void setService(Object service) {
                this.service = service;
            }
        }

        /**
         * Rethrows any throwable caught during loading of plugins
         */
        private class RethrowPluginLoadingExceptionsListener extends PluginManagerListener<P> {

            @Override
            public void pluginLoadingFailedWithException(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Throwable exception) {
                throw new RuntimeException("Plugin loader " + pluginLoader + " threw "
                        + exception.getClass().getSimpleName()
                        +" loading plugins from class loader " + classLoader
                        +" provided by " + classLoaderProvider, exception);
            }
        }
    }
}
